/*********************************************************************NVMH3****

Copyright NVIDIA Corporation 2002
TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
*AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.


Comments:

******************************************************************************/

//
// demo.cpp
//
// basic example of the use of the Cg runtime in a simple OpenGL program
//

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glext.h>
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#endif
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#ifndef CWD
#define CWD ""
#endif

typedef enum {
  S_WIRE, S_FLAT, S_SMTH, S_PHNG
} shading;

typedef struct {
  float v[3];
} normal;

typedef struct {
  float v[4];
} vertex;

typedef struct {
  float v[2];
} texel;

// globals holding the Cg context that we're storing our programs in as well
// as handles to the vertex and fragment program used in this demo
CGcontext	context;
CGprogram	vertexProgram, fragmentProgram;
CGprofile	vertexProfile, fragmentProfile;

// globals relevant to the demo object
vertex		*P=NULL;
normal		*N=NULL;
texel		*UV=NULL;
GLuint		*indices=NULL;
int		nTris;
int		nVerts;

// globals relevant to lighting
float		lightPos[3] = { 3, 2, -3 };
//float		lightPos[3] = { .3, .3, -7 };
float		lightColor[3] = { .7, .7, .7 };
float		Kd[3] = { .8, .8, .8 };
float		Ks[3] = { .9, .9, .9 };

// global rendering mode
shading		shadeMode=S_WIRE;

// global texture id
GLuint		texID;

// forward declarations of functions
int	main(int argc, char *argv[]);
void	parseCmdLine(int argc,char **argv,int *res);
void	keyboard(unsigned char key, int x, int y);
void	display();
void	handleCgError();
void	initializeGL();
void	initScene(int nu,int nv);
void	renderScene();
void	chooseCgProfiles();
void	loadCgPrograms();
void	loadCgTextures();

void parseCmdLine(int argc,char **argv,int *res)
{
	int	ch;
	int	opterr=0;
	char	*progname, *ptr;
	char	*usage =
		"usage: %s -r resolution                                   \n"
		"options:\n"
		"\t -r int      \t object resolution                       \n"
		"\t -h          \t this help message                       \n"
		;

  // get basename of argv
  if((ptr = strrchr(*argv,'/')) != NULL)	progname = ++ptr;
  else						progname = *argv;

  while((ch = getopt(argc,argv,"hr:p:"))!=EOF) {
    switch(ch) {
#ifdef __APPLE__
      // only for when user double-clicks demo.app bundle - the Finder
      // passes something like -psn_0_2228225 as argv[1] to the app
      // which we can safely ignore
      case 'p':
        fprintf(stderr,"process serial number (psn): %s\n",optarg);
        break;
#endif
      case 'r':
        (*res) = atoi(optarg);
        break;
      case 'h':
      default:
        opterr = 1;
        break;
    }
  }

  if(opterr) {
    fprintf(stderr,usage,progname);
    exit(1);
  }
}

int main(int argc, char *argv[])
{
	int	res=30;

  parseCmdLine(argc,argv,&res);

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(512, 512);
  glutCreateWindow("Cg demo");
  glutKeyboardFunc(keyboard);
  glutDisplayFunc(display);

  // basic Cg setup; register a callback function for any errors
  // and create an initial context
  cgSetErrorCallback(handleCgError);
  context = cgCreateContext();
  chooseCgProfiles();
  loadCgPrograms();
  loadCgTextures();

  initScene(res,res);

  initializeGL();

  // and all the rest happens in the display function...
  glutMainLoop();

  return 0;
}

void keyboard(unsigned char key, int x, int y)
{
  switch(key) {
    case 'q':
    case 'Q':
    case 27:
      cgDestroyProgram(vertexProgram);
      cgDestroyProgram(fragmentProgram);
      cgDestroyContext(context);
      exit(0);
      break;
    case 's':
      shadeMode = S_SMTH;
      break;
    case 'w':
      shadeMode = S_WIRE;
      break;
    case 'f':
      shadeMode = S_FLAT;
      break;
    case 'p':
      shadeMode = S_PHNG;
      break;
  }
}

void handleCgError() 
{
  fprintf(stderr, "Cg error: %s\n", cgGetErrorString(cgGetError()));
//exit(1);
}

void initializeGL()
{
//glCullFace(GL_BACK);
//glEnable(GL_CULL_FACE);	// !! WARNING !! THIS SEEMS TO SCREW UP vp/fp !!
  glClearDepth(1.0);
  glEnable(GL_DEPTH_TEST);

  // speedups
  glEnable(GL_DITHER);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

  glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightColor);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
  glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
  glMaterialf(GL_FRONT,GL_SHININESS,40.0);
  glMaterialfv(GL_FRONT,GL_SPECULAR,Ks);
  glEnable(GL_COLOR_MATERIAL);
}

void display() 
{
  // display callback function

	static int	curTime = 0;

  glDrawBuffer(GL_BACK);

  // the usual OpenGL stuff to clear the screen and set up viewing
  glClearColor(.25, .25, .25, 1.);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);                          // enable light source
  glEnable(GL_DEPTH_TEST);                      // standard depth testing
  glDepthFunc(GL_LEQUAL);
  glDepthMask(GL_TRUE);
  glDisable(GL_STENCIL_TEST);                   // no stencil testing (yet)
  glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE); // update color buffer
  glClearStencil(GL_FALSE);                     // clear stencil to 0

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(30., 1.0, .1, 100);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
//gluLookAt(4, 4, -4, 0, 0, 0, 0, 1, 0);
  gluLookAt(4, 0, -6, 0, 0, 0, 0, 1, 0);

  // make the object rotate a bit each time the display function is called
  glRotatef((float)curTime/10.0, 0, 1, 0);

  switch(shadeMode) {
    case S_WIRE:
      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,texID);
      break;
    case S_FLAT:
      glShadeModel(GL_FLAT);
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,texID);
      break;
    case S_SMTH:
      glShadeModel(GL_SMOOTH);
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,texID);
      break;
    case S_PHNG:
      glDisable(GL_LIGHTING);
      glDisable(GL_BLEND);
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

      // make sure that the vertex and fragment programs, loaded
      // in loadCgPrograms() are bound
      cgGLEnableProfile(vertexProfile);
      cgGLEnableProfile(fragmentProfile);
      cgGLBindProgram(vertexProgram);
      cgGLBindProgram(fragmentProgram);

      // we can also go ahead and bind varying parameters to vertex shader
      // that we just want to have the same value for all vertices.  the
      // vertex shader could be modified so that these were uniform for
      // better efficiency, but this gives us flexibility for the future
      cgGLSetParameter3fv(cgGetNamedParameter(vertexProgram, "diffuse"), Kd);
      cgGLSetParameter3fv(cgGetNamedParameter(vertexProgram, "specular"), Ks);

      // now bind uniform parameters to fragment shader
      cgGLSetParameter3fv(
        cgGetNamedParameter(fragmentProgram,"Plight"),lightPos);
      cgGLSetParameter3fv(
        cgGetNamedParameter(fragmentProgram,"lightColor"),lightColor);
      cgGLSetParameter1f(
        cgGetNamedParameter(fragmentProgram,"shininess"),40);

      // enable the appropriate texture for fragment shader
      cgGLEnableTextureParameter(
        cgGetNamedParameter(fragmentProgram,"diffuseMap"));

      // bind uniform parameters to vertex shader
      cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                                  "ModelViewProj"),
                                  CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                  CG_GL_MATRIX_IDENTITY);
      cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                                  "ModelView"),
                                  CG_GL_MODELVIEW_MATRIX,
                                  CG_GL_MATRIX_IDENTITY);
      cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                                  "ModelViewIT"),
                                  CG_GL_MODELVIEW_MATRIX,
                                  CG_GL_MATRIX_INVERSE_TRANSPOSE);
      break;
  }

  // and go ahead and draw the scene geometry
  renderScene();

  if(shadeMode == S_PHNG) {
    cgGLDisableTextureParameter(
      cgGetNamedParameter(fragmentProgram,"diffuseMap"));
    cgGLDisableProfile(vertexProfile);
    cgGLDisableProfile(fragmentProfile);

    glEnable(GL_LIGHTING);
  } else {
    glDisable(GL_TEXTURE_2D);
  }

  glutSwapBuffers();
  glFlush();

  ++curTime;

  // force another display(), so that the object keeps moving
  glutPostRedisplay();
}

void initScene(int nu, int nv)
{
  // geometry creation (a parameterized) sphere: cache the sphere positions,
  // normals coordinates, and vertex indices in a global array

  nTris = 2*(nu-1)*(nv-1);
  nVerts = nu*nv;

  P = new vertex[nVerts];
  N = new normal[nVerts];
  UV = new texel[nVerts];

  // fill in the position, normal, and texture coordinate arrays,
  // looping over all of the vertices, computing their parametreic
  // (u,v) coordinates, turning (u,v) coordinates into positions and
  // normals on the surface of the object

  for(int i=0, v=0; v<nv; ++v) {
    float fv = (float)v / (float)(nv-1);
    for(int u=0; u<nu; ++u) {
      float fu = (float)u / (float)(nu-1);

      UV[i].v[0] = fu;
      UV[i].v[1] = fv;

      // parametric evaluation
      float theta = M_PI * fu, phi = 2. * M_PI * fv;

      N[i].v[0] = P[i].v[0] = (float)(sin(theta) * sin(phi));
      N[i].v[1] = P[i].v[1] = (float)(sin(theta) * cos(phi));
      N[i].v[2] = P[i].v[2] = (float)(cos(theta));
                  P[i].v[3] = 1.0;
      i++;
    }
  }

  // now fill in the vertex index arrays
  indices = new GLuint[3*nTris];
  GLuint *ip = indices;
  for(int v=0; v<nv-1; ++v) {
    for(int u=0; u<nu-1; ++u) {
#define VERTEX(u, v) ((u) + (v) * nu)
      *ip++ = VERTEX(u  ,v  );
      *ip++ = VERTEX(u+1,v  );
      *ip++ = VERTEX(u+1,v+1);

      *ip++ = VERTEX(u  ,v  );
      *ip++ = VERTEX(u+1,v+1);
      *ip++ = VERTEX(u  ,v+1);
#undef VERTEX
    }
  }
}

void renderScene()
{ 
  glColor3f(1.0,1.0,1.0);

  // draw the geometry
  for(int i=0; i<3*nTris; i+=3) {
    glBegin(GL_TRIANGLES);
    for(int j=0; j<3; j++) {
      glNormal3fv(N[indices[i+j]].v);
      glTexCoord2fv(UV[indices[i+j]].v);
      glVertex4fv(P[indices[i+j]].v);
    }
    glEnd();
  }
}

void chooseCgProfiles()
{
  // choose the latest vertex and fragment profiles to use
//vertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
//fragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
//cgGLSetOptimalOptions(vertexProfile);
//cgGLSetOptimalOptions(fragmentProfile);
//fprintf(stderr, "Cg vertex profile: ");
//switch(vertexProfile) {
//  case CG_PROFILE_ARBVP1:	fprintf(stderr, "arbvp1.\n"); break;
//  case CG_PROFILE_VP30:	fprintf(stderr, "vp30.\n"); break;
//  case CG_PROFILE_VP20:	fprintf(stderr, "vp20.\n"); break;
//  default:			fprintf(stderr, "unknown.\n"); break;
//}
//fprintf(stderr, "Cg fragment profile: ");
//switch(fragmentProfile) {
//  case CG_PROFILE_ARBFP1:	fprintf(stderr, "arbfp1.\n"); break;
//  case CG_PROFILE_FP30:	fprintf(stderr, "fp30.\n"); break;
//  case CG_PROFILE_FP20:	fprintf(stderr, "fp20.\n"); break;
//  default:			fprintf(stderr, "unknown.\n"); break;
//}

//return;

  // choose the vertex and fragment profiles to use.  try to use
  // CG_PROFILE_ARBVFP1 and CG_PROFILE_ARBFP1, depending on hardware support.
  // if those aren't available, fall back to CG_PROFILE_VP30 and
  // CG_PROFILE_FP30, respectively.

  // make sure that the appropriate profiles are available on the user's system.
  if (cgGLIsProfileSupported(CG_PROFILE_ARBVP1)) {
    vertexProfile = CG_PROFILE_ARBVP1;
    fprintf(stderr, "Cg vertex profile: arbvp1.\n");
  // try VP30
  } else if (cgGLIsProfileSupported(CG_PROFILE_VP30)) {
    vertexProfile = CG_PROFILE_VP30;
    fprintf(stderr, "Cg vertex profile: vp30.\n");
  // try VP20
  } else if (cgGLIsProfileSupported(CG_PROFILE_VP20)) {
    vertexProfile = CG_PROFILE_VP20;
    fprintf(stderr, "Cg vertex profile: vp20.\n");
  } else {
    fprintf(stderr,"Neither arbvp1, vp30, or vp20 vertex profiles supported");
    fprintf(stderr," on this system.\n");
    exit(1);
  }

  if (cgGLIsProfileSupported(CG_PROFILE_ARBFP1)) {
    fragmentProfile = CG_PROFILE_ARBFP1;
    fprintf(stderr, "Cg fragment profile: arbfp1.\n");
  // try FP30
  } else if (cgGLIsProfileSupported(CG_PROFILE_FP30)) {
    fragmentProfile = CG_PROFILE_FP30;
    fprintf(stderr, "Cg fragment profile: fp30.\n");
  // try FP20
  } else if (cgGLIsProfileSupported(CG_PROFILE_FP20)) {
    fragmentProfile = CG_PROFILE_FP20;
    fprintf(stderr, "Cg fragment profile: fp20.\n");
  } else {
    fprintf(stderr,"Neither arbfp1, fp30, or fp20 fragment profiles supported");
    fprintf(stderr," on this system.\n");
    exit(1);
  }
}

void loadCgPrograms()
{
  assert(cgIsContext(context));

  // load and compile the vertex program from demo_vert.cg; hold on to the
  // handle to it that is returned
  fprintf(stderr, "Cg vertex program creating on this system.\n");
  vertexProgram = cgCreateProgramFromFile(context,
                                          CG_SOURCE,
                                          CWD "demo_vert.cg",
                                          vertexProfile, NULL, NULL);
  if(!cgIsProgramCompiled(vertexProgram)) {
    fprintf(stderr, "Cg vertex program compiling on this system.\n");
    cgCompileProgram(vertexProgram);
    fprintf(stderr, "Cg vertex program compiled on this system.\n");
  }

  // enable the appropriate vertex profile and load the vertex program
  cgGLEnableProfile(vertexProfile);
  fprintf(stderr, "Cg vertex program loading on this system.\n");
  cgGLLoadProgram(vertexProgram);

  // and similarly set things up for the fragment program
  fprintf(stderr, "Cg fragment program creating on this system.\n");
  fragmentProgram = cgCreateProgramFromFile(context,
                                            CG_SOURCE,
                                            CWD "demo_frag.cg",
                                            fragmentProfile, NULL, NULL);
  if(!cgIsProgramCompiled(fragmentProgram)) {
    fprintf(stderr, "Cg fragment program compiling on this system.\n");
    cgCompileProgram(fragmentProgram);
    fprintf(stderr, "Cg fragment program compiled on this system.\n");
  }

  cgGLEnableProfile(fragmentProfile);
  fprintf(stderr, "Cg fragment program loading on this system.\n");
  cgGLLoadProgram(fragmentProgram);

  cgGLDisableProfile(vertexProfile);
  cgGLDisableProfile(fragmentProfile);
}

void loadCgTextures()
{

  // there is only one texture needed here--we'll set up a basic
  // checkerboard--which is used to modulate the diffuse channel in the
  // fragment shader
  glGenTextures(1, &texID);

  // Basic OpenGL texture state setup
  glBindTexture(GL_TEXTURE_2D, texID);
  glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS,GL_TRUE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);

    // fill in the texture map
#define RES 512
  float *data = new float[RES*RES*4];
  float *dp = data;
  for(int i = 0; i < RES; ++i) {
    for(int j = 0; j < RES; ++j) {
      if((i/32+j/32) % 2) {
        *dp++ = .7;
        *dp++ = .7;
        *dp++ = .7;
      } else {
        *dp++ = .1;
        *dp++ = .1;
        *dp++ = .1;
      }
      *dp++ = 1.;
    }
  }

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,RES,RES,0,GL_RGBA,GL_FLOAT,data);
  delete [] data;

  // tell Cg which texture id should be associated with the sampler2D
  // parameter to the fragment shader
  cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram, "diffuseMap"),
                          texID);
}
